/* pager.c -- The display program of sawfish.wm.ext.pager
   $Id: pager.c,v 1.4 2002/04/06 15:01:14 pfeiffer Exp $

   Copyright (C) 2002 Daniel Pfeiffer <occitan@esperanto.org>
   Copyright (C) 2000 Satyaki Das <satyaki@theforce.stanford.edu>
                      Hakon Alstadheim

   This file is part of sawfish.

   sawfish is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   sawfish is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with sawfish; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <gtk/gtk.h>

#ifdef USE_SAWFISH_LIBCLIENT
/* 
 * @SD: Should include libclient.h instead.
 */
int client_open (char *display);
char *client_eval(char *form, int *lenp, int *errorp);
void client_close (void);
#endif /* USE_SAWFISH_LIBCLIENT */

enum { bg, hilit_bg, win, focus_win, win_border, vp_divider, ws_divider, vp_frame };

GdkColor colors[ws_divider+1];
GdkGC *gc[ws_divider+2];
GdkColormap *cmap = NULL;
GdkPixmap *pixmap = NULL;
GtkWidget *drawing_area;
GtkWidget *window;
GdkRectangle update_rect;

/* can display this many, including all instances of stickies */
#define MAXWINDOWS 512
struct {
  long id, x, y, width, height;
} windows[MAXWINDOWS], *lowest, *w, *dragged_window, *tooltip_window;

int width, height, pwidth, pheight, wwidth, wheight;
int offset_x, offset_y;
int vp_x, vp_y;
long focus_id;
int mouse_x = -1;
int mouse_y = -1;

gint delete_event(GtkWidget *, GdkEvent *, gpointer);
gint destroy_event(GtkWidget *, GdkEvent *, gpointer);
gint configure_event(GtkWidget *, GdkEvent *, gpointer);
gint leave_notify_event(GtkWidget *, GdkEvent *, gpointer);
gint expose_event(GtkWidget *, GdkEventExpose *, gpointer);
gint button_press_event(GtkWidget *, GdkEventButton *, gpointer);
gint button_release_event(GtkWidget *, GdkEventButton *, gpointer);
gint motion_notify_event(GtkWidget *, GdkEventMotion *, gpointer);
gint wait_stdin(gpointer, gint, GdkInputCondition);
void draw_pager(GtkWidget *);
gint draw_tooltip();
char parse_stdin(void);
void send_command(char *cmd);



int main(int argc, char *argv[])
{
  GtkWidget *vbox;
  gtk_init(&argc, &argv);

#ifdef USE_SAWFISH_LIBCLIENT
  if (client_open (NULL))
    exit(1);
  atexit(client_close);
#endif /* USE_SAWFISH_LIBCLIENT */

  /* Get the dimensions and colors of the pager and viewport and focus */
  parse_stdin();
  parse_stdin();
  parse_stdin();
  parse_stdin();

  update_rect.x = 0;
  update_rect.y = 0;

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), vbox);

  drawing_area = gtk_drawing_area_new();
  gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), width, height);
  gtk_box_pack_start(GTK_BOX(vbox), drawing_area, FALSE, FALSE, 0);

  /* Signals to quit */
  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
		     GTK_SIGNAL_FUNC(delete_event), NULL);
  gtk_signal_connect(GTK_OBJECT(window), "destroy",
		     GTK_SIGNAL_FUNC(destroy_event), NULL);

  /* Wait for input from standard input */
  gdk_input_add(0, GDK_INPUT_READ, GTK_SIGNAL_FUNC(wait_stdin), drawing_area);

  /* Change the viewport when a button is pressed */
  gtk_signal_connect(GTK_OBJECT(drawing_area), "motion_notify_event",
		     (GtkSignalFunc) motion_notify_event, NULL);
  gtk_signal_connect(GTK_OBJECT(drawing_area), "button_press_event",
		     (GtkSignalFunc) button_press_event, NULL);
  gtk_signal_connect(GTK_OBJECT(drawing_area), "button_release_event",
		     (GtkSignalFunc) button_release_event, NULL);
  gtk_signal_connect(GTK_OBJECT(drawing_area), "leave_notify_event",
		     (GtkSignalFunc) leave_notify_event, NULL);
  gtk_widget_set_events(drawing_area, GDK_EXPOSURE_MASK
			            | GDK_LEAVE_NOTIFY_MASK
			            | GDK_BUTTON_PRESS_MASK
                                    | GDK_BUTTON_RELEASE_MASK
		                    | GDK_POINTER_MOTION_MASK
		                    | GDK_POINTER_MOTION_HINT_MASK
			            | GDK_LEAVE_NOTIFY_MASK);

  /* Initialize and draw the pixmap */
  gtk_signal_connect(GTK_OBJECT(drawing_area), "expose_event",
		     (GtkSignalFunc) expose_event, NULL);
  gtk_signal_connect(GTK_OBJECT(drawing_area), "configure_event",
		     (GtkSignalFunc) configure_event, NULL);

  gtk_widget_show(drawing_area);
  gtk_widget_show(vbox);
  gtk_widget_show(window);
  gtk_main();
  return 0;
}

void send_command(char *cmd)
{
#ifdef USE_SAWFISH_LIBCLIENT
  int rv;
  client_eval(cmd,NULL,&rv);
#else
  char cmdline[2048];
  sprintf(cmdline, "sawfish-client -e \"%s\"", cmd);
  system(cmdline);
#endif /* USE_SAWFISH_LIBCLIENT */
}

gint find_w(int x, int y) {
  for (w = windows; w <= lowest; w++)
    if (w->x <= x && w->x + w->width > x &&
	w->y <= y && w->y + w->height > y)
      return TRUE;
  return FALSE;
}



gint expose_event(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
  gdk_draw_pixmap(widget->window,
		  widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		  pixmap,
		  0, 0, 0, 0,
		  width, height);
  return TRUE;
}

gint configure_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  int i;
  if (pixmap) {
    gdk_pixmap_unref(pixmap);
    for (i = bg; i <= vp_frame; i++)
      gdk_gc_unref(gc[i]);
  }

  pixmap = gdk_pixmap_new(widget->window, width, height, -1);
  for (i = bg; i <= ws_divider; i++)
    gdk_gc_set_foreground(gc[i] = gdk_gc_new(pixmap), &colors[i]);

  gdk_gc_set_foreground(gc[vp_frame] = gdk_gc_new(pixmap), &colors[vp_divider]);
  gdk_gc_set_line_attributes(gc[vp_frame], 1, GDK_LINE_ON_OFF_DASH,
			     GDK_CAP_BUTT, GDK_JOIN_BEVEL);

  draw_pager(widget);

  return TRUE;
}

gint leave_notify_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  mouse_x = width;
  mouse_y = height;

  draw_tooltip();
  return TRUE;
}  

gint motion_notify_event(GtkWidget *widget, GdkEventMotion *event, gpointer d)
{
  int x, y;
  GdkModifierType state;

  if (event->is_hint)
    gdk_window_get_pointer(event->window, &x, &y, &state);
  else {
    x = event->x;
    y = event->y;
    state = event->state;
  }

  mouse_x = x;
  mouse_y = y;

  if (!dragged_window && draw_tooltip())
    return TRUE;

  if (state & GDK_BUTTON3_MASK) {
    draw_pager(widget);

    if (dragged_window)
      gdk_draw_rectangle(pixmap, widget->style->black_gc, FALSE,
			 x-offset_x, y-offset_y,
			 dragged_window->width, dragged_window->height);
    gtk_widget_draw(widget, &update_rect);
  }

  return TRUE;
}    

gint button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer d)
{
  char cmd[64];
  int x, y;

  mouse_x = x = (int) event->x;
  mouse_y = y = (int) event->y;

  /* Button1 changes viewport */
  if (event->button == 1) {
    sprintf(cmd, "(pager-goto %ld %d %d)", find_w(x, y) ? w->id : 0, x, y);
    send_command(cmd);
    draw_tooltip();
  }

  /* Button2 raises/lowers current window */
  else if (event->button == 2 && find_w(x, y)) {
    sprintf(cmd, "(pager-change-depth %ld)", w->id);
    send_command(cmd);
    draw_tooltip();
  }

  /* Button3 is for dragging the selected window */
  else if (event->button == 3 && find_w(x, y)) {
    draw_tooltip();
    dragged_window = w;

    /* Offset of top left from cursor */
    offset_x = x - w->x;
    offset_y = y - w->y;

    /* move the selected window out of sight */
    w->x = width;
    w->y = height;

    draw_pager(widget);
    gdk_draw_rectangle(pixmap, widget->style->black_gc, FALSE,
		       x-offset_x, y-offset_y,
		       dragged_window->width, dragged_window->height);
    gtk_widget_draw(widget, &update_rect);
  }
  return TRUE;
}

gint button_release_event(GtkWidget *widget, GdkEventButton *event, gpointer d)
{
  int x,y;

  if (event->button == 3) {
    x = (int) event->x;
    y = (int) event->y;

    if (dragged_window) {
      char cmd[64];
      sprintf(cmd, "(pager-move-window %ld %d %d)",
	      dragged_window->id, x-offset_x, y-offset_y);
      send_command(cmd);
      dragged_window = NULL;
    }
    draw_pager(widget);
    gtk_widget_draw(widget, &update_rect);
  }

  return TRUE;
}



void draw_pager(GtkWidget *widget)
{
  int i;

  /* make the whole thing color background */
  gdk_draw_rectangle(pixmap, gc[bg], TRUE, 1, 1, width, height);

  /* highlight the current viewport with color hilit background */
  gdk_draw_rectangle(pixmap, gc[hilit_bg], TRUE,
		     vp_x, vp_y, pwidth, pheight);

  /* draw the boundaries of the different viewports */
  for (i=pwidth; i<width; i+=pwidth)
    gdk_draw_line(pixmap, gc[vp_divider], i, 1, i, height-1);
  for (i=pheight; i<height; i+=pheight)
    gdk_draw_line(pixmap, gc[vp_divider], 1, i, width-1, i);

  /* draw the windows */
  for (w = lowest; w >= windows; w--) {
    gdk_draw_rectangle(pixmap, gc[(w->id == focus_id) ? focus_win : win], TRUE,
		       w->x+1, w->y+1, w->width-2, w->height-2);
    gdk_draw_rectangle(pixmap, gc[win_border], FALSE,
		       w->x, w->y, w->width-1, w->height-1);
  }

  /* draw the boundaries of the different workspaces */
  for (i=0; i<=width; i+=wwidth)
    gdk_draw_line(pixmap, gc[ws_divider], i, 0, i, height);
  for (i=0; i<=height; i+=wheight)
    gdk_draw_line(pixmap, gc[ws_divider], 1, i, width, i);

  /* frame the current viewport above all else in case it's covered */
  gdk_draw_rectangle(pixmap, gc[vp_frame], FALSE,
		     vp_x, vp_y, pwidth, pheight);

  gtk_widget_draw(widget, &update_rect);

  return;
}

gint draw_tooltip()
{
  if(find_w(mouse_x, mouse_y)) {
    char cmd[64];
    if (w == tooltip_window)
      return TRUE;
    tooltip_window = w;
    sprintf(cmd, "(pager-tooltip %ld)", w->id);
    send_command(cmd);
    return TRUE;
  } else if (tooltip_window)
    send_command("(pager-tooltip)");
  tooltip_window = NULL;
  return FALSE;
}



int last;
long get_number()
{
  char buffer;
  long number = 0;
  int sign = 1;
  while (read (0, &buffer, 1) == 1) {
    if (buffer >= '0' && buffer <= '9') {
      number = 10 * number + buffer - '0';
    } else if (buffer == ' ' || buffer == '\n') {
      if (sign == -1) {
	number = -number;
	sign = 1;
      }
      last = (buffer == '\n');
      return number;
    } else if (buffer == '-') {
      sign = -1;
    }
  }
  return number;		/* never reached */
}

char parse_stdin()
{
  char type;
  long i;
  int repeat;

  /* This reads the type of the data:
   *   W - regular info about what to display
   *   s - size info to resize the window
   *   c - color change
   *   w - info about one window
   *   f - focus change
   */
  read(0, &type, 1);
  switch (type) {
  case 'f':
    focus_id = get_number();
    break;
  case 'v':
    vp_x = get_number();
    vp_y = get_number();
    break;
  case 'W':
    last = 0;
    lowest = windows;
    lowest->id = get_number();	/* maybe no number */
    if (last)
      lowest--;
    while (!last) {
      if (lowest <= windows + MAXWINDOWS) {
	lowest->x = get_number();
	lowest->y = get_number();
	lowest->width = get_number();
	lowest->height = get_number();
	if (lowest->width && lowest->height && !last)
	  lowest++;
	if (!last)
	  lowest->id = get_number();
      }
      else
	get_number();
    }
    break;
  case 'w':
    i = get_number();
    for (w = lowest; w >= windows; w--) {
      if (w->id == i) {
	w->x = get_number();
	w->y = get_number();
	w->width = get_number();
	w->height = get_number();
	break;
      }
    }
    break;
  case 's':
    pwidth = get_number();
    pheight = get_number();
    wwidth = get_number();
    wheight = get_number();
    width = update_rect.width = get_number() + 1;
    height = update_rect.height = get_number() + 1;
    break;
  case 'c':
    repeat = (cmap != NULL);
    cmap = gdk_colormap_get_system();
    for (i = bg; i <= ws_divider; i++) {
      if (repeat)
	gdk_colormap_free_colors(cmap, &colors[i], 1);
      colors[i].red = get_number();
      colors[i].green = get_number();
      colors[i].blue = get_number();
      gdk_color_alloc(cmap, &colors[i]);
    }
  }
  return type;
}

gint wait_stdin(gpointer data, gint source, GdkInputCondition cond)
{

  if (parse_stdin() != 's') {
    draw_pager(data);
    draw_tooltip();
  }
  else {
    /* Initialize the picture */
    gdk_window_resize(window->window, width, height);
    gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), width, height);
  }

  return TRUE;
}

gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  return FALSE;
}

gint destroy_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  gtk_main_quit();
  return FALSE;
}
